"""This file contains constants used by img2catalog"""

from pathlib import Path

CONFIG_HOME_PATH = Path.home() / ".img2catalog" / "config.toml"

XNAT_HOST_ENV = "XNAT_HOST"
XNATPY_HOST_ENV = "XNATPY_HOST"
XNAT_USER_ENV = "XNAT_USER"
XNAT_PASS_ENV = "XNAT_PASS"

FDP_SERVER_ENV = "IMG2CATALOG_FDP"
FDP_USER_ENV = "IMG2CATALOG_FDP_USER"
FDP_PASS_ENV = "IMG2CATALOG_FDP_PASS"

SPARQL_ENV = "IMG2CATALOG_SPARQL_ENDPOINT"

# Default setting
REMOVE_OPTIN_KEYWORD = True



import logging

import pytest

logger = logging.getLogger(__name__)

####
# Below code is taken from XNATPy, under the Apache 2.0 license
# (C) BIGR, ErasmusMC
# Included as xnat.tests isn't default
from typing import Any, Pattern, Union
from unittest.mock import patch

import requests
from pytest_mock import MockerFixture
from requests import Response
from requests_mock import Mocker
from xnat.session import XNATSession

from img2catalog.xnat_parser import XNATParserError, xnat_private_project


class CreatedObject:
    def __init__(self, uri, type_, fieldname, **kwargs):
        self.uri = uri
        self.type = type_
        self.fieldname = fieldname
        self.kwargs = kwargs


class XnatpyRequestsMocker(Mocker):
    def request(self, method: str, url: Union[str, Pattern[str]], **kwargs: Any) -> Response:
        url = f"https://xnat.example.com/{url.lstrip('/')}"
        return super().request(method, url, **kwargs)


@pytest.fixture(scope="function")
def xnatpy_mock() -> XnatpyRequestsMocker:
    with XnatpyRequestsMocker() as mocker:
        yield mocker


@pytest.fixture(scope="function")
def xnatpy_connection(mocker: MockerFixture, xnatpy_mock: XnatpyRequestsMocker) -> XNATSession:
    # Create a working mocked XNATpy connection object
    threading_patch = mocker.patch("xnat.session.threading")  # Avoid background threads getting started
    # logger = logging.getLogger('xnatpy_test')

    xnatpy_mock.get("/data/JSESSION")
    xnatpy_mock.delete("/data/JSESSION")
    xnatpy_mock.get("/data/version", status_code=404)
    xnatpy_mock.get(
        "/xapi/siteConfig/buildInfo",
        json={
            "version": "1.7.5.6",
            "buildNumber": "1651",
            "buildDate": "Tue Aug 20 18:10:41 CDT 2019",
            "sha": "5696414138",
            "isDirty": "false",
            "commit": "2",
            "tag": "1.7.5.4",
            "shaFull": "5696414138d8c95288bf45c8eac2150ba041e867",
            "branch": "master",
            "timestamp": "1566342641000",
        },
    )
    requests_session = requests.Session()

    # Set cookie for JSESSION/timeout
    cookie = requests.cookies.create_cookie(
        domain="xnat.example.com",
        name="JSESSIONID",
        value="3EFD012EF2FA60EF44BA72ED5925F074",
    )
    requests_session.cookies.set_cookie(cookie)

    cookie = requests.cookies.create_cookie(
        domain="xnat.example.com",
        name="SESSION_EXPIRATION_TIME",
        value='"1668081619871,900000"',
    )
    requests_session.cookies.set_cookie(cookie)

    xnat_session = XNATSession(
        server="https://xnat.example.com",
        logger=logger,
        interface=requests_session,
    )

    # Patch create object to avoid a lot of hassle
    def create_object(uri, type_=None, fieldname=None, **kwargs):
        return CreatedObject(uri, type_, fieldname, **kwargs)

    xnat_session.create_object = create_object

    yield xnat_session

    # Close connection before the mocker gets cleaned
    xnat_session.disconnect()

    # Clean mocker
    xnatpy_mock.reset()

    # Stop patch of threading
    mocker.stop(threading_patch)


######


@patch("xnat.core.XNATBaseObject")
@pytest.mark.parametrize(
    "test_input, expected",
    [
        ("public", False),
        ("PUBLIC", False),
        ("protected", False),
        ("ProTected", False),
        ("private", True),
        ("Private", True),
    ],
)
def test_private_project_capitalization(
    project, test_input, expected, xnatpy_mock: Mocker, xnatpy_connection: XNATSession
):
    xnatpy_mock.get("/data/projects/example/accessibility", text=test_input)
    project.xnat_session = xnatpy_connection
    project.name = "example"
    project.uri = "/data/projects/example"

    assert xnat_private_project(project) == expected


@patch("xnat.core.XNATBaseObject")
def test_private_project_invalid_value(project, xnatpy_mock: Mocker, xnatpy_connection: XNATSession):
    xnatpy_mock.get("/data/projects/example/accessibility", text="random_invalid_string_output")
    project.xnat_session = xnatpy_connection
    project.name = "example"
    project.uri = "/data/projects/example"

    with pytest.raises(XNATParserError):
        xnat_private_project(project)



from unittest.mock import patch

import pytest

from img2catalog.xnat_parser import _check_optin_optout, split_keywords


@pytest.mark.parametrize(
    "test_str, expected",
    [
        # Various type of whitespace and variations
        ("", []),
        ("        ", []),
        (" .  , ;  .. .,;:", []),
        # Keywords with different amounts of space inbetween
        ("xnat python", ["xnat", "python"]),
        ("  xnat      python ", ["xnat", "python"]),
        # Below is real example. The official XNAT specification is not followed by the writer
        ("artificial intelligence, radiomics", ["artificial", "intelligence", "radiomics"]),
        # Some random punctuations and stuff
        ("artificial;intelligence.   radiomics", ["artificial", "intelligence", "radiomics"]),
        ("artificial ; intelligence.,.,.,radiomics", ["artificial", "intelligence", "radiomics"]),
    ],
)
def test_keyword_splitter(test_str, expected):
    output_list = split_keywords(test_str)

    # We compare it as a set as we do not care about the order (the RDF serializer mangles it anyways)
    assert set(expected) == set(output_list)


@patch("xnat.core.XNATBaseObject")
def test_no_optin_optout(project):
    project.keywords = "test demo optout_keyword"
    config = {"img2catalog": {}}

    # First test no config
    assert _check_optin_optout(project, config)


@patch("xnat.core.XNATBaseObject")
def test_optout(project):
    # project without keywords
    config = {"img2catalog": {"optout": "optout_keyword"}}
    assert _check_optin_optout(project, config)

    project.keywords = "test demo optout_keyword"
    assert not _check_optin_optout(project, config)


@patch("xnat.core.XNATBaseObject")
def test_optin(project):
    config = {"img2catalog": {"optin": "optin_keyword"}}
    project.keywords = "test demo optin_keyword"

    assert _check_optin_optout(project, config)

    project.keywords = "test demo"
    assert not _check_optin_optout(project, config)


@patch("xnat.core.XNATBaseObject")
def test_optin_noconfig(project):
    config = {}
    project.keywords = "test demo"

    assert _check_optin_optout(project, config)



import pathlib
import sys
from unittest.mock import ANY, Mock, patch

import pytest
from rdflib import DCAT, DCTERMS, Graph, URIRef
from rdflib.compare import to_isomorphic
from sempyro.dcat.dcat_dataset import DCATDataset
from sempyro.vcard import VCARD

from img2catalog.cli_app import cli_click, load_img2catalog_configuration
from img2catalog.const import (
    XNAT_HOST_ENV,
    XNAT_PASS_ENV,
    XNAT_USER_ENV,
    XNATPY_HOST_ENV,
    SPARQL_ENV,
    FDP_USER_ENV,
    FDP_PASS_ENV,
    FDP_SERVER_ENV,
)

TEST_CONFIG = pathlib.Path(__file__).parent / "example-config.toml"


@pytest.fixture()
def empty_graph():
    graph = Graph()
    graph.bind("dcat", DCAT)
    graph.bind("dcterms", DCTERMS)
    graph.bind("v", VCARD)
    return graph


@pytest.fixture()
def toml_patch_target():
    # Python 3.11 and up has tomllib built-in, for 3.10 and lower we use tomli which provides
    # the same functonality. We check if it's Python 3.10 or lower to patch the correct target.
    if sys.version_info < (3, 11):
        return "tomli.load"
    else:
        return "tomllib.load"


@pytest.fixture()
def dummy_dcat_dataset():
    d = DCATDataset(title=["test project"], description=["test description"])
    return d


@patch("xnat.connect")
@patch("img2catalog.cli_app.xnat_to_RDF")
def test_cli_connect(xnat_to_RDF, connect, empty_graph, isolated_cli_runner):
    # Mock context manager of xnatpy and the XNAT to RDF function
    connect.__enter__.return_value = True
    xnat_to_RDF.return_value = empty_graph

    # Run isolated (to keep log files safe)
    result = isolated_cli_runner.invoke(cli_click, ["--server", "http://example.com", "--verbose", "dcat"])

    connect.assert_called_once_with(server="http://example.com", user=None, password=None)
    xnat_to_RDF.assert_called_once()

    assert result.exit_code == 0


@patch("xnat.connect")
@patch("img2catalog.cli_app.xnat_to_RDF")
def test_example_cli(xnat_to_RDF, connect, empty_graph, isolated_cli_runner):
    # Mock context manager of xnatpy and the XNAT to RDF function
    connect.__enter__.return_value = True
    xnat_to_RDF.return_value = empty_graph

    # Run isolated (to keep log files safe)
    result = isolated_cli_runner.invoke(cli_click, ["--verbose", "-s", "http://example.com", "dcat"])

    connect.assert_called_once_with(server="http://example.com", user=None, password=None)
    xnat_to_RDF.assert_called_once()

    assert result.exit_code == 0

    # assert False


@patch("xnat.connect")
@patch("img2catalog.cli_app.xnat_to_RDF")
def test_anonymous_envhost(xnat_to_RDF, connect, empty_graph, isolated_cli_runner, monkeypatch):
    # Mock context manager of xnatpy and the XNAT to RDF function
    connect.__enter__.return_value = True
    xnat_to_RDF.return_value = empty_graph

    monkeypatch.setenv(XNATPY_HOST_ENV, "http://test.example.com")
    monkeypatch.setenv(XNAT_HOST_ENV, "http://fail_test.example.com")

    # Run isolated (to keep log files safe)
    result = isolated_cli_runner.invoke(cli_click, ["dcat"])

    connect.assert_called_once_with(server="http://test.example.com", user=None, password=None)
    xnat_to_RDF.assert_called_once()

    assert result.exit_code == 0


@patch("xnat.connect")
@patch("img2catalog.cli_app.xnat_to_RDF")
def test_second_env_var(xnat_to_RDF, connect, empty_graph, isolated_cli_runner, monkeypatch):
    # Mock context manager of xnatpy and the XNAT to RDF function
    connect.__enter__.return_value = True
    xnat_to_RDF.return_value = empty_graph

    # monkeypatch.setenv(XNATPY_HOST_ENV, "http://test.example.com")
    monkeypatch.setenv(XNAT_HOST_ENV, "http://pass_test.example.com")
    # Run isolated (to keep log files safe)
    result = isolated_cli_runner.invoke(cli_click, ["dcat"])

    connect.assert_called_once_with(server="http://pass_test.example.com", user=None, password=None)
    xnat_to_RDF.assert_called_once()

    assert result.exit_code == 0


# @pytest.mark.xfail(reason="Clearing password not implemented yet")
@patch("xnat.connect")
@patch("img2catalog.cli_app.xnat_to_RDF")
def test_user_pass_prio_env(xnat_to_RDF, connect, empty_graph, isolated_cli_runner, monkeypatch):
    # Mock context manager of xnatpy and the XNAT to RDF function
    connect.__enter__.return_value = True
    xnat_to_RDF.return_value = empty_graph

    monkeypatch.setenv(XNAT_USER_ENV, "fail_user")
    monkeypatch.setenv(XNAT_PASS_ENV, "fail_password")
    # monkeypatch.setenv(XNAT_HOST_ENV, "http://fail_test.example.com")
    # Run isolated (to keep log files safe)
    result = isolated_cli_runner.invoke(cli_click, ["-u", "pass_user", "-s", "http://test.example.com", "dcat"])

    # FIXME Not sure if this is desired behavior. Ideally, if the username is set as an argument,
    # it should prompt for the password or at least ignore the environment variable.
    connect.assert_called_once_with(server="http://test.example.com", user="pass_user", password=None)
    xnat_to_RDF.assert_called_once()

    assert result.exit_code == 0


@patch("xnat.connect")
@patch("img2catalog.cli_app.xnat_to_RDF")
def test_user_pass_envvar(xnat_to_RDF, connect, empty_graph, isolated_cli_runner, monkeypatch):
    # Mock context manager of xnatpy and the XNAT to RDF function
    connect.__enter__.return_value = True
    xnat_to_RDF.return_value = empty_graph

    monkeypatch.setenv(XNAT_USER_ENV, "pass_user")
    monkeypatch.setenv(XNAT_PASS_ENV, "password")
    # Run isolated (to keep log files safe)
    result = isolated_cli_runner.invoke(
        cli_click,
        ["-s", "http://test.example.com", "dcat"],
    )

    connect.assert_called_once_with(server="http://test.example.com", user="pass_user", password="password")
    xnat_to_RDF.assert_called_once()

    assert result.exit_code == 0


@patch("xnat.connect")
@patch("img2catalog.cli_app.xnat_to_RDF")
@pytest.mark.parametrize(
    "test_input, expected",
    [
        (["-s", "http://test.example.com", "dcat"], {"format": "turtle"}),
        (
            [
                "-s",
                "http://test.example.com",
                "dcat",
                "-o",
                "tester_1.ttl",
                "-f",
                "xml",
            ],
            {"destination": "tester_1.ttl", "format": "xml"},
        ),
        (
            [
                "-s",
                "http://test.example.com",
                "dcat",
                "-o",
                "tester_2.ttl",
            ],
            {"destination": "tester_2.ttl", "format": "turtle"},
        ),
    ],
)
def test_serialize_cli_args(xnat_to_RDF, connect, test_input, expected, empty_graph, isolated_cli_runner):
    # Mock context manager of xnatpy and the XNAT to RDF function
    connect.__enter__.return_value = True
    xnat_to_RDF.return_value = empty_graph

    # Run isolated (to keep log files clean)
    with patch.object(empty_graph, "serialize") as serializer:
        result = isolated_cli_runner.invoke(
            cli_click,
            test_input,
        )
        serializer.assert_called_once_with(**expected)

    connect.assert_called_once_with(server="http://test.example.com", user=None, password=None)
    xnat_to_RDF.assert_called_once()

    assert result.exit_code == 0


@patch("xnat.connect")
def test_nonexisting_config(connect, isolated_cli_runner):
    result = isolated_cli_runner.invoke(cli_click, ["-s", "http://example.com", "--config", "non_existing_file.toml"])

    assert not connect.called, "Function was called despite having to error out"

    # Click says that a UsageError should have exit code 2.
    assert result.exit_code == 2


def test_config_loader_error():
    config_path = Mock(spec=pathlib.Path)
    config_path.exists.return_value = False

    with pytest.raises(FileNotFoundError):
        load_img2catalog_configuration(config_path)


@pytest.mark.parametrize("config_param", [None, TEST_CONFIG])
@patch("img2catalog.configmanager.CONFIG_HOME_PATH", TEST_CONFIG)
@patch("builtins.open")
def test_config_dir(fileopen, toml_patch_target, config_param):
    with patch(toml_patch_target) as load:
        load_img2catalog_configuration(config_param)
        # Make sure the correct configuration is loaded
        fileopen.assert_called_once_with(TEST_CONFIG, "rb")
        # Make sure the file de-serializer is called, not the string de-serializer
        load.assert_called_once()


@patch("img2catalog.cli_app.xnat_to_FDP")
@patch("fairclient.fdpclient.FDPClient.__init__")
@patch("xnat.connect")
def test_fdp_cli(connect, mock_FDPClient, xnat_to_FDP, isolated_cli_runner):
    connect.__enter__.return_value = True

    mock_FDPClient.return_value = None

    isolated_cli_runner.invoke(
        cli_click,
        [
            "--verbose",
            "-s",
            "http://example.com",
            "fdp",
            "--fdp",
            "http://fdp.example.com",
            "-u",
            "test",
            "-p",
            "more_test",
            "-c",
            "http://catalog.example.com",
        ],
    )

    connect.assert_called_once_with(server="http://example.com", user=ANY, password=ANY)
    xnat_to_FDP.assert_called_once()
    pass


@patch("img2catalog.cli_app.xnat_to_FDP")
@patch("fairclient.fdpclient.FDPClient.__init__")
@patch("fairclient.sparqlclient.FDPSPARQLClient.__init__")
@patch("xnat.connect")
def test_fdp_cli_env(connect, mock_SPARQLClient, mock_FDPClient, xnat_to_FDP, isolated_cli_runner, monkeypatch):
    connect.__enter__.return_value = True

    mock_FDPClient.return_value = None
    mock_SPARQLClient.return_value = None

    monkeypatch.setenv(XNAT_HOST_ENV, "http://example.com")
    monkeypatch.setenv(FDP_USER_ENV, "userFDP")
    monkeypatch.setenv(FDP_PASS_ENV, "passwordFDP")
    monkeypatch.setenv(FDP_SERVER_ENV, "http://fdp.example.com")
    monkeypatch.setenv(SPARQL_ENV, "http://sparql.example.com")

    isolated_cli_runner.invoke(
        cli_click,
        [
            "--verbose",
            "fdp",
            "-c",
            "http://catalog.example.com",
        ],
    )

    mock_FDPClient.assert_called_once_with("http://fdp.example.com", "userFDP", "passwordFDP")
    mock_SPARQLClient.assert_called_once_with(URIRef("http://sparql.example.com"))
    connect.assert_called_once_with(server="http://example.com", user=ANY, password=ANY)
    xnat_to_FDP.assert_called_once()
    pass


@patch("xnat.connect")
@patch("img2catalog.cli_app.xnat_to_DCATDataset")
def test_output_project(
    xnat_to_DCATDataset,
    connect,
    dummy_dcat_dataset,
    empty_graph,
    isolated_cli_runner,
):
    # patch the session.projects such that it returns the id it was called with
    connect.return_value.__enter__.return_value.projects.__getitem__.side_effect = lambda x: x

    # Always return a mock DCATDataset object and URI
    xnat_to_DCATDataset.return_value = (dummy_dcat_dataset, URIRef("http://example.com"))

    # with patch.object(dummy_dcat_dataset, "to_graph", return_value=empty_graph) as serializer:
    result = isolated_cli_runner.invoke(
        cli_click,
        ["--verbose", "-s", "http://example.com", "project", "test_project"],
    )

    result_graph = empty_graph.parse(result.stdout_bytes, format="ttl")
    reference_graph = empty_graph.parse(source=pathlib.Path(__file__).parent / "references" / "mock_dataset.ttl")

    # Verify known output
    assert to_isomorphic(reference_graph) == to_isomorphic(result_graph)

    connect.assert_called_once_with(server="http://example.com", user=ANY, password=ANY)
    xnat_to_DCATDataset.assert_called_with("test_project", ANY)
    connect.return_value.__enter__.return_value.projects.__getitem__.assert_called_once_with("test_project")


@patch("xnat.connect")
@patch("img2catalog.cli_app.xnat_to_DCATDataset")
def test_output_project_file(
    xnat_to_DCATDataset,
    connect,
    dummy_dcat_dataset,
    empty_graph,
    isolated_cli_runner,
):
    # patch the session.projects such that it returns the id it was called with
    connect.return_value.__enter__.return_value.projects.__getitem__.side_effect = lambda x: x

    # Always return a mock DCATDataset object and URI
    xnat_to_DCATDataset.return_value = (dummy_dcat_dataset, URIRef("http://example.com"))

    isolated_cli_runner.invoke(
        cli_click,
        ["--verbose", "-s", "http://example.com", "project", "test_project", "-o", "test_project.xml", "-f", "xml"],
    )

    result_graph = empty_graph.parse(source="test_project.xml", format="xml")
    reference_graph = empty_graph.parse(source=pathlib.Path(__file__).parent / "references" / "mock_dataset.ttl")

    # Verify known output
    assert to_isomorphic(reference_graph) == to_isomorphic(result_graph)

    # Make sure right functions are called
    connect.assert_called_once_with(server="http://example.com", user=ANY, password=ANY)
    xnat_to_DCATDataset.assert_called_with("test_project", ANY)
    connect.return_value.__enter__.return_value.projects.__getitem__.assert_called_once_with("test_project")



import logging
import sys
from logging import FileHandler, StreamHandler
from logging.handlers import RotatingFileHandler
from os import PathLike
from typing import Union

LOGGING_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
LOGGING_DATEFMT = "%Y-%m-%d %H:%M:%S"


class Logger:
    """Logging class that logs Warnings to stderr and Info to file.

    Parameters
    ----------
    logger_name : str
        Internal name of logger.
    logger_path : str, PathLike
        Path to the logfile

    """

    def __init__(self, logger_name: str, logger_path: Union[str, PathLike] = None):
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.INFO)

        # output to stderr as to allow clean piping of output
        console_handler = StreamHandler(sys.stderr)
        console_handler.setLevel(logging.WARNING)

        # file_handler = FileHandler(f'./{logger_name}.log')
        # file_handler.setLevel(logging.INFO)
        self.formatter = logging.Formatter(LOGGING_FORMAT, datefmt=LOGGING_DATEFMT)
        # file_handler.setFormatter(formatter)
        console_handler.setFormatter(self.formatter)

        logger.addHandler(console_handler)
        # logger.addHandler(file_handler)
        self.logger = logger
        self.logger_path = logger_path

    def _add_file_handler(self, logger_path: Union[str, PathLike] = None):
        """Adds a file handler to logging

        Parameters
        ----------
        logger_path : Union[str, PathLike], optional
            Location to store logfile. If not set, will use self.logger_path, if that isn't set
            either, a logfile with the name of the logger will be stored in the current directory.
        """
        if logger_path is not None:
            self.logger_path = logger_path
        if self.logger_path is None:
            self.logger_path = f"./{self.logger_name}.log"

        file_handler = RotatingFileHandler(self.logger_path, maxBytes=1e6, backupCount=3)
        file_handler.setLevel(logging.INFO)

        file_handler.setFormatter(self.formatter)
        self.logger.addHandler(file_handler)

    def setLevel(self, loglevel):
        """Sets loglevel for both the logfile as well as stderr output

        Parameters
        ----------
        loglevel : str, Int
            Logging level, must be str or int.
        """
        self.logger.setLevel(loglevel)
        if self.logger.handlers:
            for handler in self.logger.handlers:
                handler.setLevel(loglevel)
        else:
            logging.basicConfig(level=loglevel)



import logging
from pathlib import Path
from typing import Dict

# Python < 3.11 does not have tomllib, but tomli provides same functionality
try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

from img2catalog.const import CONFIG_HOME_PATH

logger = logging.getLogger(__name__)


def example_standard_config() -> str:
    """This function returns a serialized version of the example configuration file.

    While not the cleanest solution, it does allow for easy packaging.
    In the future, this could be replaced by a configuration migration function.

    Returns
    -------
    str
        String of a toml-serialized configuration file.
    """
    standard_config = """[img2catalog]
# optin = "include_catalog"
# optout = "exclude_catalog"

[dataset]
# Theme should be Health, according to the EU Data Theme vocabulary
theme = "http://publications.europa.eu/resource/authority/data-theme/HEAL"

# Default License for datasets is "available upon reasonable request"
# I can't find that, so putting a placeholder here
license = "http://example.com/license#nolicense"

[dataset.contact_point]
# override = true
full_name = "Example Data Management office"
email = "datamanager@example.com"

[dataset.publisher]
name = ["Example publisher list",]
identifier = "http://example.com"

[catalog]
title = "Example XNAT catalog"
description = "This is an example XNAT catalog description"

[catalog.publisher]
name = ["Example publishing institution",]
identifier = "http://www.example.com/institution#example"

[distribution.default]
title = "XNAT imaging distribution"
description = "Link to XNAT instance where the imaging data can be accessed."
"""
    return standard_config


def load_img2catalog_configuration(config_path: Path = None) -> Dict:
    """Loads a configuration file for img2catalog

    First, it checks if config_path is given. If not, it will look for ~/.img2catalog/config.toml,
    if that file also doesn't exist it will load an example configuration from the project rootfolder.

    Parameters
    ----------
    config_path : Path, optional
        Path to configuration file to load, by default None

    Returns
    -------
    Dict
        Dictionary with loaded configuration properties

    Raises
    ------
    FileNotFoundError
        If config_path is specified yet does not exist
    """
    if config_path:
        if not config_path.exists():
            raise FileNotFoundError(f"Configuration file does not exist at {config_path}")
    elif (config_path := CONFIG_HOME_PATH).exists():
        pass
    else:
        logger.warning("No configuration file found or specified! Using example configuration")
        config = tomllib.loads(example_standard_config())
        return config

    logger.info("Using configuration file %s", config_path)

    with open(config_path, "rb") as f:
        config = tomllib.load(f)

    return config



"""Simple tool to query an XNAT instance and serialize projects as datasets"""

import datetime
import html
import logging
import re
from typing import Dict, List, Tuple, Union

from fairclient.fdpclient import FDPClient
from fairclient.sparqlclient import FDPSPARQLClient
from fairclient.utils import add_or_update_dataset, remove_node_from_graph
from rdflib import DCAT, DCTERMS, FOAF, RDF, Graph, URIRef
from sempyro.foaf import Agent
from sempyro.hri_dcat.hri_catalog import HRICatalog
from sempyro.hri_dcat.hri_dataset import HRIDataset
from sempyro.vcard import VCARD, VCard
from tqdm import tqdm
from xnat.core import XNATBaseObject
from xnat.session import XNATSession

from img2catalog.const import REMOVE_OPTIN_KEYWORD

logger = logging.getLogger(__name__)


class XNATParserError(ValueError):
    """Exception that can contain an list of errors from the XNAT parser.

    Parameters
    ----------
    message: str
        Exception message
    error_list: list
        List of strings containing error messages. Default is None

    """

    def __init__(self, message: str, error_list: List[str] = None):
        super().__init__(message)
        self.error_list = error_list


def xnat_to_DCATDataset(project: XNATBaseObject, config: Dict) -> Tuple[HRIDataset, URIRef]:
    """This function populates a DCAT Dataset class from an XNat project

    Currently fills in the title, description and keywords. The first two are mandatory fields
    for a dataset in DCAT-AP, the latter is a bonus.

    Note: XNAT sees keywords as one string field, this function assumes they are comma-separated.

    Parameters
    ----------
    project : XNatListing
        An XNat project instance which is to be generated
    config : Dict
        A dictionary containing the configuration of img2catalog

    Returns
    -------
    HRIDataset
        HRIDataset object with fields filled in
    URIRef
        Subject that could be used
    """
    # Specification from XNAT:  Optional: Enter searchable keywords. Each word, separated by a space,
    # can be used independently as a search string.
    keywords = filter_keyword(split_keywords(project.keywords), config)

    error_list = []
    if not (project.pi.firstname or project.pi.lastname):
        error_list.append("Cannot have empty name of PI")
    if not project.description:
        error_list.append("Cannot have empty description")

    if error_list:
        raise XNATParserError("Errors encountered during the parsing of XNAT.", error_list=error_list)

    project_uri = project.external_uri()

    creator_list = list([xnat_investigator_to_Agent(project.pi)])

    # FIXME
    # Workaround for xnatpy issue #68
    # https://gitlab.com/radiology/infrastructure/xnatpy/-/issues/68
    if project.investigators:
        for i in range(len(project.investigators)):
            creator = xnat_investigator_to_Agent(project.investigators[i])
            creator_list.append(creator)

    publisher_foaf = [Agent(**config["dataset"]["publisher"])]

    dataset_license = URIRef(config["dataset"]["license"])
    themes = [URIRef(config["dataset"]["theme"])]

    # TODO These are stub values, should be modified to reflect something slightly more accurate
    issued = datetime.datetime.now()
    modified = datetime.datetime.now()

    project_description = html.unescape(project.description)

    dataset_dict = {
        "title": [project.name],
        "description": [project_description],
        "creator": creator_list,
        "keyword": keywords,
        "identifier": project_uri,
        "license": dataset_license,
        "publisher": publisher_foaf,
        "theme": themes,
        "issued": issued,
        "modified": modified,
    }

    contact_point_vcard = [contact_point_vcard_from_config(config)]
    if any(contact_point_vcard):
        dataset_dict["contact_point"] = contact_point_vcard

    project_dataset = HRIDataset(**dataset_dict)

    return project_dataset, URIRef(project_uri)


def xnat_investigator_to_Agent(investigator) -> Agent:
    creator_foaf = Agent(
        name=[f"{investigator.title or ''} {investigator.firstname} {investigator.lastname}".strip()],
        identifier="http://example.com",  # Should be ORCID?
    )

    return creator_foaf


def xnat_to_DCATCatalog(session: XNATSession, config: Dict) -> HRICatalog:
    """Creates a DCAT-AP compliant Catalog from XNAT instance

    Parameters
    ----------
    session : XNATSession
        An XNATSession of the XNAT instance that is going to be queried
    config : Dict
        A dictionary containing the configuration of img2catalog

    Returns
    -------
    HRICatalog
        HRICatalog object with fields populated
    """
    publisher_foaf = [Agent(**config["catalog"]["publisher"])]

    catalog = HRICatalog(
        title=[config["catalog"]["title"]],
        description=[config["catalog"]["description"]],
        publisher=publisher_foaf,
    )
    return catalog


def xnat_to_RDF(session: XNATSession, config: Dict) -> Graph:
    """Creates a DCAT-AP compliant Catalog of Datasets from XNAT

    Parameters
    ----------
    session : XNATSession
        An XNATSession of the XNAT instance that is going to be queried
    config : Dict
        A dictionary containing the configuration of img2catalog

    Returns
    -------
    Graph
        An RDF graph containing DCAT-AP
    """
    export_graph = Graph()

    # To make output cleaner, bind these prefixes to namespaces
    export_graph.bind("dcat", DCAT)
    export_graph.bind("dcterms", DCTERMS)
    export_graph.bind("foaf", FOAF)
    export_graph.bind("vcard", VCARD)

    catalog = xnat_to_DCATCatalog(session, config)
    catalog_uri = URIRef(session.url_for(session))

    dataset_list = xnat_list_datasets(session, config)

    # Assign an empty list so we can easily append datasets
    catalog.dataset = []

    for dcat_dataset, subject in dataset_list:
        d = dcat_dataset.to_graph(subject)
        export_graph += d
        catalog.dataset.append(subject)

    export_graph += catalog.to_graph(catalog_uri)

    return export_graph


def xnat_to_FDP(
    session: XNATSession, config: Dict, catalog_uri: URIRef, fdpclient: FDPClient, sparqlclient: FDPSPARQLClient = None
) -> None:
    """Pushes DCAT-AP compliant Datasets to FDP

    Parameters
    ----------
    session : XNATSession
        An XNATSession of the XNAT instance that is going to be queried
    config : Dict
        A dictionary containing the configuration of img2catalog
    catalog_uri : URIRef
        URI of the catalog in which the datasets will be placed
    fdpclient : FDPClient
        An instance of FDPClient of the FDP where the datasets will be put
    sparqlclient : FDPSPARQLClient, optional
        An instance of FDPSPARQLClient which can be used for updating existing datasts, by default None

    Returns
    -------
    Graph
        An RDF graph containing DCAT-AP
    """
    dataset_list = xnat_list_datasets(session, config)

    for dataset, subject in tqdm(dataset_list):
        dataset_graph = dataset.to_graph(subject)

        # This is FDP specific: Dataset points back to the Catalog
        dataset_graph.add((subject, DCTERMS.isPartOf, catalog_uri))

        logger.debug("Going to push %s to FDP", dataset.title)
        try:
            add_or_update_dataset(dataset_graph, fdpclient, dataset.identifier, catalog_uri, sparqlclient)
        except Exception as e:
            logger.warning("Error pushing dataset to FDP: %s", e)


def xnat_list_datasets(session: XNATSession, config: Dict) -> List[HRIDataset]:
    """Acquires a list of elligible XNAT datasets

    Parameters
    ----------
    session : XNATSession
        An XNATSession of the XNAT instance that is going to be queried
    config : Dict
        A dictionary containing the configuration of img2catalog

    Returns
    -------
    List[DCATDataset, URI]
        List of DCAT models of elligible datasets, along with their URIs (for subjects)

    """
    failure_counter = 0
    dataset_list = []

    for p in tqdm(session.projects.values()):
        try:
            if not _check_elligibility_project(p, config):
                logger.debug("Project %s not elligible, skipping", p.id)
                continue

            dcat_dataset = xnat_to_DCATDataset(p, config)
            dataset_list.append(dcat_dataset)

        except XNATParserError as v:
            logger.info(f"Project {p.name} could not be converted into DCAT: {v}")

            for err in v.error_list:
                logger.info(f"- {err}")
            failure_counter += 1
            continue

    if failure_counter > 0:
        logger.warning("There were %d projects with invalid data for DCAT generation", failure_counter)

    return dataset_list


def filter_keyword(xnat_keywords: Union[List[str], None], config: Dict) -> List[str]:
    """Filters the opt-in keyword from the xnat keywords list

    If no opt-in keyword is set, all keywords will be returned.
    If remove_keywords is set to False, all keywords will be returned.
    If the opt-in keyword cannot be found, all keywords will be returned.

    Parameters
    ----------
    xnat_keywords : Union[List[str], None]
        List of XNAT keywords
    config : Dict
        img2catalog configuration dictionary

    Returns
    -------
    List[str]
        List of XNAT keywords, with the opt-in keyword filtered out if necessary
    """
    try:
        optin_kw = config["img2catalog"].get("optin")
    except KeyError:
        # If key not found, means config is not set, so no opt-in/opt-out set so all are included
        return xnat_keywords

    remove_keyword = config["img2catalog"].get("remove_optin", REMOVE_OPTIN_KEYWORD)

    if remove_keyword:
        if optin_kw in xnat_keywords:
            xnat_keywords.remove(optin_kw)
            return xnat_keywords

    return xnat_keywords


def split_keywords(xnat_keywords: Union[str, None]) -> List[str]:
    """Takes an XNAT keyword list and splits it up into a list of keywords

    Removes all non alphanumeric characters from the keywords

    Parameters
    ----------
    xnat_keywords : str
        String containing all keywords. Space, comma or (semi-)colon-separated.

    Returns
    -------
    List[str]
        List of keywords, empty list if there are no keywords
    """
    keyword_list = []
    if xnat_keywords:
        if len(xnat_keywords.strip()) > 0:
            keyword_list = [
                kw.strip()
                for kw in re.split(r"[\.,;: ]", xnat_keywords)
                # for kw in xnat_keywords.strip().replace(".", " ").replace(",", " ").replace(";", " ").split(" ")
            ]

    # Filter out all empty strings, then return list
    return list(filter(None, keyword_list))


def xnat_private_project(project) -> bool:
    """This function checks if an XNAT project is a private project

    Parameters
    ----------
    project : XNAT ProjectData
        The project of which the permission status needs to be investigated

    Returns
    --------
    bool
        Returns True if the project is private, False if it is protected or public

    Raises
    ------
    XNATParserError
        If the XNAT API returns an unknown value for accessibility
    """
    # The API URI is documented as part of the XNAT Project Attributes API at
    # https://wiki.xnat.org/xnat-api/project-attributes-api
    # As it is not exposed as part of the XNAT XSD, XNATPy does not generate a field for it

    # The API documentation says it should be Title case, in practice XNAT returns lowercase
    # Therefore I consider the case to be unreliable
    accessibility = project.xnat_session.get(f"{project.uri}/accessibility").text.casefold()
    known_accesibilities = ["public", "private", "protected"]
    if accessibility.casefold() not in known_accesibilities:
        raise XNATParserError(f"Unknown permissions of XNAT project: accessibility is '{accessibility}'")

    if accessibility == "private".casefold():
        return True
    else:
        return False


def _check_optin_optout(project, config: Dict) -> bool:
    """This function checks if the project is elligible for indexing, given the opt-in/opt-out keywords

    Parameters
    ----------
    project : XNAT
        XNAT project of which the elligiblity needs to be determined
    config : Dict
        Configuration dictionary with the opt-in/opt-out keys

    Returns
    -------
    bool
        Returns True if a project is elligible for indexing, False if it is not.
    """
    try:
        optin_kw = config["img2catalog"].get("optin")
        optout_kw = config["img2catalog"].get("optout")
    except KeyError:
        # If key not found, means config is not set, so no opt-in/opt-out set so always elligible.
        return True

    if optin_kw:
        if optin_kw not in split_keywords(project.keywords):
            logger.debug("Project %s does not contain keyword on opt-in list, skipping", project)
            return False
    elif optout_kw:
        if optout_kw in split_keywords(project.keywords):
            logger.debug("Project %s contains keyword on opt-out list, skipping", project)
            return False

    return True


def _check_elligibility_project(project, config: Dict) -> bool:
    """Checks if a project is elligible for indexing given its properties and img2catalog config

    Parameters
    ----------
    p : XNAT Project
        The XNATpy project to be index4ed
    config : Dict
        Dictionary containing img2catalog config

    Returns
    -------
    bool
        Returns True if the project could be indexed, False if not.
    """
    # Check if project is private. If it is, skip it
    if xnat_private_project(project):
        logger.debug("Project %s is private, not elligible", project.id)
        return False

    if not _check_optin_optout(project, config):
        logger.debug("Skipping project %s due to keywords", project.id)
        return False

    logger.debug("Project %s is elligible for indexing", project)

    return True


def contact_point_vcard_from_config(config: Dict) -> Union[VCard, None]:
    """Generates a VCard for contact_point of datasets

    Parameters
    ----------
    config : Dict
        img2catalog configuration

    Returns
    -------
    Union[VCARD, None]
        VCard if info is present in configuration, None if not.
    """
    try:
        contact_config = config["dataset"]["contact_point"]
    except KeyError:
        return None

    # email should be URIRef (but not path, https://stackoverflow.com/a/27151227)
    contact_email = contact_config.get("email")
    if contact_email:
        if not contact_email.startswith("mailto:"):
            contact_email = f"mailto:{contact_email}"
        contact_email = URIRef(contact_email)

        contact_vcard = VCard(
            full_name=[contact_config["full_name"]], hasEmail=[contact_email], hasUID=URIRef("http://example.com")
        )

        return contact_vcard
    else:
        return None



import pathlib
from typing import Any, Dict
from unittest.mock import MagicMock, patch

import pytest
from sempyro.dcat.dcat_catalog import DCATCatalog
from sempyro.dcat.dcat_dataset import DCATDataset

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

import xnat
from freezegun import freeze_time
from rdflib import DCAT, DCTERMS, Graph, URIRef
from rdflib.compare import to_isomorphic

from img2catalog.xnat_parser import (
    VCARD,
    XNATParserError,
    _check_elligibility_project,
    filter_keyword,
    xnat_list_datasets,
    xnat_to_DCATDataset,
    xnat_to_FDP,
    xnat_to_RDF,
)

TEST_CONFIG = pathlib.Path(__file__).parent / "example-config.toml"


# Taken from cedar2fdp
@pytest.fixture()
def empty_graph():
    graph = Graph()
    graph.bind("dcat", DCAT)
    graph.bind("dcterms", DCTERMS)
    graph.bind("v", VCARD)
    return graph


@pytest.fixture()
def config():
    """Loads the default configuration TOML"""
    config_path = TEST_CONFIG

    with open(config_path, "rb") as f:
        config = tomllib.load(f)

    return config


@pytest.fixture()
def mock_catalog():
    catalog = DCATCatalog(title=["Example XNAT catalog"], description=["This is an example XNAT catalog description"])
    return catalog


@pytest.fixture()
def mock_dataset():
    dataset = DCATDataset(title=["test project"], description=["test description"])
    return dataset


@patch("xnat.session.BaseXNATSession")
def test_empty_xnat(session, empty_graph: Graph, config: Dict[str, Any]):
    """Test case for an XNAT with no projects at all"""
    # XNATSession is a key-value store so pretend it is a Dict
    session.projects = {}
    session.url_for.return_value = "https://example.com"

    empty_graph = empty_graph.parse(source="tests/references/empty_xnat.ttl")

    expected = xnat_to_RDF(session, config)

    # Compare to reference graph
    assert to_isomorphic(expected) == to_isomorphic(empty_graph)


@freeze_time("2024-04-01")
@patch("xnat.core.XNATBaseObject")
def test_valid_project_no_investigator(project, empty_graph: Graph, config: Dict[str, Any]):
    """Test if a valid project generates valid output"""
    project.name = "Basic test project to test the img2catalog"
    project.description = "In this project, we test xnat and dcat and make sure a description appears."
    project.external_uri.return_value = "http://localhost/data/archive/projects/test_img2catalog"
    project.keywords = "test demo dcat"
    project.pi.firstname = "Albus"
    project.pi.lastname = "Dumbledore"
    project.pi.title = "prof."
    project.investigators = False

    empty_graph = empty_graph.parse(source="tests/references/valid_project_no_investigator.ttl")
    dcat, uri = xnat_to_DCATDataset(project, config)
    gen = dcat.to_graph(uri)

    assert to_isomorphic(empty_graph) == to_isomorphic(gen)


@freeze_time("2024-04-01")
@patch("xnat.core.XNATBaseObject")
def test_valid_project(project, empty_graph: Graph, config: Dict[str, Any]):
    """Test if a valid project generates valid output"""
    project.name = "Basic test project to test the img2catalog"
    project.description = "In this project, we test &quot;xnat&quot; &amp; dcat and make sure a description appears."
    project.external_uri.return_value = "http://localhost/data/archive/projects/test_img2catalog"
    project.keywords = "test demo dcat"
    project.pi.firstname = "Albus"
    project.pi.lastname = "Dumbledore"
    project.pi.title = "prof."
    # project
    project.investigators.__len__.return_value = 1
    project.investigators[0].firstname = "Minerva"
    project.investigators[0].lastname = "McGonagall"
    project.investigators[0].title = "Prof."

    empty_graph = empty_graph.parse(source="tests/references/valid_project.ttl")
    dcat, uri = xnat_to_DCATDataset(project, config)
    gen = dcat.to_graph(uri)

    assert to_isomorphic(empty_graph) == to_isomorphic(gen)


@patch("xnat.core.XNATBaseObject")
def test_empty_description(project, config: Dict[str, Any]):
    """Test if a valid project generates valid output"""
    project.name = "Basic test project to test the img2catalog"
    project.description = None
    project.external_uri.return_value = "http://localhost/data/archive/projects/test_img2catalog"
    project.keywords = "test demo dcat"
    project.pi.firstname = "Albus"
    project.pi.lastname = "Dumbledore"
    project.pi.title = "prof."

    with pytest.raises(ValueError):
        xnat_to_DCATDataset(project, config).to_graph(userinfo_format=VCARD.VCard)


@patch("xnat.core.XNATBaseObject")
def test_invalid_pi(project, config: Dict[str, Any]):
    """Make sure if PI field is invalid, an exception is raised"""
    project.name = "Basic test project to test the img2catalog"
    project.description = "In this project, we test xnat and dcat and make sure a description appears."
    project.external_uri.return_value = "http://localhost/data/archive/projects/test_img2catalog"
    project.keywords = "test demo dcat"
    project.pi.firstname = None
    project.pi.lastname = None

    with pytest.raises(ValueError):
        xnat_to_DCATDataset(project, config)


@freeze_time("2024-04-01")
@patch("xnat.core.XNATBaseObject")
def test_no_keywords(project, empty_graph: Graph, config: Dict[str, Any]):
    """Valid project without keywords, make sure it is not defined in output"""
    project.name = "Basic test project to test the img2catalog"
    project.description = "In this project, we test xnat and dcat and make sure a description appears."
    project.external_uri.return_value = "http://localhost/data/archive/projects/test_img2catalog"
    project.keywords = ""
    project.pi.firstname = "Albus"
    project.pi.lastname = "Dumbledore"
    project.pi.title = "prof."

    empty_graph = empty_graph.parse(source="tests/references/no_keyword.ttl")
    dcat, uri = xnat_to_DCATDataset(project, config)
    gen = dcat.to_graph(uri)

    assert to_isomorphic(empty_graph) == to_isomorphic(gen)


@pytest.mark.parametrize(
    "private, optin, expected", [(False, True, True), (True, True, False), (False, False, False), (True, False, False)]
)
@patch("xnat.core.XNATBaseObject")
@patch("img2catalog.xnat_parser._check_optin_optout")
@patch("img2catalog.xnat_parser.xnat_private_project")
def test_project_elligiblity(xnat_private_project, _check_optin_optout, project, private, optin, expected):
    project.__str__.return_value = "test project"
    project.id = "test"
    xnat_private_project.return_value = private
    _check_optin_optout.return_value = optin

    assert _check_elligibility_project(project, None) == expected
    # pass


@pytest.mark.parametrize(
    "config, expected",
    [
        ({}, ["apple", "banana", "pear"]),
        ({"img2catalog": {"optin": "banana"}}, ["apple", "pear"]),  # test default remove
        ({"img2catalog": {}}, ["apple", "banana", "pear"]),  # test no settings
        ({"img2catalog": {"optin": "banana", "remove_optin": False}}, ["apple", "banana", "pear"]),  # do not remove
        (
            {"img2catalog": {"optin": "orange", "remove_optin": True}},
            ["apple", "banana", "pear"],
        ),  # Should never happen but still
    ],
)
def test_keyword_filter(config, expected):
    keyword_list = ["apple", "banana", "pear"]
    assert sorted(filter_keyword(keyword_list, config)) == expected


@patch("img2catalog.xnat_parser._check_elligibility_project")
@patch("img2catalog.xnat_parser.xnat_to_DCATDataset")
def test_xnat_lister(xnat_to_DCATDataset, _check_elligibility_project):
    class SimpleProject:
        def __init__(self, project_id):
            self.id = None
            self.name = project_id

    # xnat_list_datasets
    session = MagicMock(spec=xnat.session.XNATSession)
    session.projects.values.return_value = [
        SimpleProject("p1"),
        SimpleProject("p2"),
        SimpleProject("p3"),
        SimpleProject("p4"),
    ]

    _check_elligibility_project.side_effect = [True, False, True, True]
    xnat_to_DCATDataset.side_effect = [
        ("project_1", "uri1"),
        ("project_3", "uri2"),
        XNATParserError("Test error", ["This project cannot be converted"]),
    ]

    list_result = xnat_list_datasets(session, {})

    assert list_result == [("project_1", "uri1"), ("project_3", "uri2")]

    assert _check_elligibility_project.call_count == 4
    assert xnat_to_DCATDataset.call_count == 3


@freeze_time("2024-04-01")
@patch("xnat.session.BaseXNATSession")
@patch("img2catalog.xnat_parser.xnat_to_DCATCatalog")
@patch("img2catalog.xnat_parser.xnat_list_datasets")
def test_xnat_to_rdf(xnat_list_datasets, xnat_to_DCATCatalog, session, mock_dataset, mock_catalog, config, empty_graph):
    """Tests of XNAT to RDF pushing happens with correct arguments and graph is modified correctly"""
    xnat_to_DCATCatalog.return_value = mock_catalog

    session.projects = {}
    session.url_for.return_value = "https://example.com"

    xnat_list_datasets.return_value = [(mock_dataset, URIRef("https://example.com/dataset"))]

    result_graph = xnat_to_RDF(session, config)
    reference_graph = empty_graph.parse(source="tests/references/minimal_dcat_catalog_dataset.ttl")

    assert to_isomorphic(result_graph) == to_isomorphic(reference_graph)


@patch("img2catalog.xnat_parser.add_or_update_dataset")
@patch("img2catalog.xnat_parser.xnat_list_datasets")
def test_xnat_to_fdp_push(xnat_list_datasets, add_or_update_dataset, mock_dataset, config, empty_graph):
    """Tests of XNAT to RDF pushing happens even when errors happen"""

    xnat_list_datasets.return_value = [(mock_dataset, URIRef("http://example.com/dataset"))]

    xnat_to_FDP(None, config, URIRef("http://example.com/catalog"), None, None)

    # Check if function is called and correct term added to Graph
    add_or_update_dataset.assert_called_once()
    assert (
        URIRef("http://example.com/dataset"),
        DCTERMS.isPartOf,
        URIRef("http://example.com/catalog"),
    ) in add_or_update_dataset.call_args.args[0], "FDP catalog reference missing"

    pass


@patch("img2catalog.xnat_parser.add_or_update_dataset")
@patch("img2catalog.xnat_parser.xnat_list_datasets")
def test_xnat_to_fdp_push_error(xnat_list_datasets, add_or_update_dataset, mock_dataset, config, empty_graph):
    xnat_list_datasets.return_value = [
        (mock_dataset, URIRef("http://example.com/dataset1")),
        (mock_dataset, URIRef("http://example.com/dataset2")),
    ]
    add_or_update_dataset.side_effect = [ValueError, None]

    xnat_to_FDP(None, config, URIRef("http://example.com/catalog"), None, None)

    assert add_or_update_dataset.call_count == 2
    assert (
        URIRef("http://example.com/dataset2"),
        DCTERMS.isPartOf,
        URIRef("http://example.com/catalog"),
    ) in add_or_update_dataset.call_args_list[1].args[0], "FDP catalog reference missing"

    pass



import logging
from pathlib import Path

import click
import xnat
from click_option_group import MutuallyExclusiveOptionGroup, optgroup
from fairclient.fdpclient import FDPClient
from fairclient.sparqlclient import FDPSPARQLClient
from rdflib import URIRef

from img2catalog import log
from img2catalog.__about__ import __version__
from img2catalog.configmanager import load_img2catalog_configuration

# from xnat.client.helpers import xnatpy_login_options, connect_cli
from img2catalog.const import (
    FDP_PASS_ENV,
    FDP_SERVER_ENV,
    FDP_USER_ENV,
    XNAT_HOST_ENV,
    XNAT_PASS_ENV,
    XNAT_USER_ENV,
    XNATPY_HOST_ENV,
    SPARQL_ENV,
)
from img2catalog.xnat_parser import xnat_to_DCATDataset, xnat_to_FDP, xnat_to_RDF

logger = logging.getLogger(__name__)


def __connect_xnat(server: str, username, password):
    """This function collects credentials and connects to XNAT

    Parameters
    ----------
    server : str
        XNAT server to connect (including https://)
    args : Namespace
        Namespace containing commandline arguments

    Returns
    -------
    XNATSession
    """
    # if not (server):
    #     if not (server := os.environ.get(XNATPY_HOST_ENV)):
    #         if not (server := os.environ.get(XNAT_HOST_ENV)):
    #             raise RuntimeError("No server specified: no argument nor environment variable found")
    # if not (username):
    #     if not (username := os.environ.get(XNAT_USER_ENV)):
    #         logger.info("No username set, using anonymous/netrc login")
    # if not (password):
    #     if not (password := os.environ.get(XNAT_PASS_ENV)):
    #         logger.info("No password set, using anonymous/netrc login")

    logger.debug("Connecting to server %s using username %s", server, username)

    session = xnat.connect(server=server, user=username, password=password)

    return session


def cli_main():
    # try:
    cli_click()
    # except Exception as e:
    #     print(f"Error running img2catalog:\n{e}")
    #     exit(-1)


# @click.command(name='dcat', help="Export XNAT to DCAT")
# @optgroup.group('Server configuration', help='The configuration of the XNAT server', required=False)
@click.group(invoke_without_command=True)
# wontfix
# https://github.com/pallets/click/issues/714
@click.option(
    "-s",
    "--server",
    type=str,
    envvar=[XNATPY_HOST_ENV, XNAT_HOST_ENV],
    required=True,
    help=f"URI of the server to connect to (including http:// or https://). If not set, will use environment variables {XNATPY_HOST_ENV} or {XNAT_HOST_ENV}.",
)
@click.option(
    "-u",
    "--username",
    type=str,
    default=None,
    envvar=XNAT_USER_ENV,
    help=f"Username to use, leave empty to use netrc entry or anonymous login or environment variable {XNAT_USER_ENV}.",
)
@click.option(
    "-p",
    "--password",
    type=str,
    default=None,
    envvar=XNAT_PASS_ENV,
    help=(
        "Password to use with the username, leave empty when using netrc. If a"
        " username is given and no password or environment variable, there will be a prompt on the console"
        f" requesting the password. Environment variable: {XNAT_PASS_ENV}"
    ),
)
@click.option(
    "-c",
    "--config",
    default=None,
    type=click.Path(exists=True, path_type=Path, readable=True),
    help="Configuration file to use. If not set, will use ~/.img2catalog/config.toml if it exists.",
)
@click.option("-v", "--verbose", is_flag=True, default=False, help="Enables debugging mode.")
@click.option(
    "-l",
    "--logfile",
    default="./img2catalog.log",
    type=click.Path(file_okay=True, dir_okay=False, path_type=Path, writable=True),
    help="Path of logfile to use. Default is img2catalog.log in current directory",
)
@optgroup(cls=MutuallyExclusiveOptionGroup)
@optgroup.option(
    # Both opt-in and opt-out at the same time is not very logical, so it is not allowed.
    "--optin",
    type=str,
    help="Opt-in keyword. If set, only projects with this keyword will be included",
    default=None,
)
@optgroup.option(
    "--optout", type=str, help="Opt-out keyword. If set, projects with this keyword will be excluded", default=None
)
# @xnatpy_login_options
@click.pass_context
@click.version_option(__version__)
def cli_click(
    ctx: click.Context,
    server: str,
    username: str,
    password: str,
    config: click.Path,
    verbose: bool,
    logfile: click.Path,
    optin: str,
    optout: str,
    **kwargs,
):
    """This tool queries metadata from an XNAT server"""
    ctx.ensure_object(dict)
    log._add_file_handler(logfile)
    logger.info("======= img2catalog New Run ========")
    if verbose:
        log.setLevel(logging.DEBUG)
        logger.debug("Verbose mode enabled")

    config = load_img2catalog_configuration(config)

    if optin or optout:
        config["img2catalog"]["optin"] = optin
        config["img2catalog"]["optout"] = optout

    # with connect_cli(cli=False, **kwargs) as session:
    # If username is not environment variable and password is, that's usually not intended
    # Thus we clear password so xnatpy can deal with it
    if ctx.get_parameter_source("username") != click.core.ParameterSource.ENVIRONMENT:
        if ctx.get_parameter_source("password") == click.core.ParameterSource.ENVIRONMENT:
            password = None

    ctx.obj["xnat_conn"] = __connect_xnat(server, username, password)
    ctx.obj["config"] = config

    # output_dcat(server, username, password, output, format, config)


@cli_click.command(name="dcat")
@click.option(
    "-o",
    "--output",
    "output",
    default=None,
    type=click.Path(writable=True, dir_okay=False),
    help="Destination file to write output to. If not set, the script will print serialized output to stdout.",
)
@click.option(
    "-f",
    "--format",
    default="turtle",
    type=click.Choice(
        ["xml", "n3", "turtle", "nt", "pretty-xml", "trix", "trig", "nquads", "json-ld", "hext"], case_sensitive=False
    ),
    help=(
        "The format that the output should be written in. This value references a"
        " Serializer plugin in RDFlib. Supportd values are: "
        ' "xml", "n3", "turtle", "nt", "pretty-xml", "trix", "trig", "nquads",'
        ' "json-ld" and "hext". Defaults to "turtle".'
    ),
)
@click.pass_context
def output_dcat(ctx: click.Context, output: click.Path, format: str):
    # , server, username, password, output, format, config):
    config = ctx.obj["config"]
    with ctx.obj["xnat_conn"] as session:
        logger.debug("Connected to XNAT server")
        g = xnat_to_RDF(session, config)
        logger.debug("Finished acquiring RDF graph")

    if output:
        logger.debug("Output option set, serializing output to file %s in %s format", output, format)
        g.serialize(destination=output, format=format)

    else:
        logger.debug("Sending output to stdout")
        print(g.serialize(format=format))


@click.option("-f", "--fdp", envvar=FDP_SERVER_ENV, type=str, required=True, help="URL of FDP to push datasets to")
@click.option("-u", "--username", envvar=FDP_USER_ENV, type=str, required=True, help="Username of FDP to push to")
@click.option("-p", "--password", envvar=FDP_PASS_ENV, type=str, required=True, help="Password of FDP to push to")
@click.option(
    "-c", "--catalog", default=None, type=URIRef, help="Catalog URI where datasets will be placed in", required=True
)
@click.option(
    "-s",
    "--sparql",
    envvar=SPARQL_ENV,
    type=URIRef,
    help=" URL of SPARQL endpoint of FDP, used for querying which dataset to update",
)
@cli_click.command(name="fdp")
@click.pass_context
def output_fdp(ctx: click.Context, fdp: str, username: str, password: str, catalog: URIRef, sparql: str):
    config = ctx.obj["config"]

    # For some reason, in testing, execution doesn't progress beyond this line
    fdpclient = FDPClient(fdp, username, password)

    if sparql:
        sparqlclient = FDPSPARQLClient(sparql)
    else:
        sparqlclient = None

    if not catalog:
        if not (catalog := config["img2catalog"]["fdp"]["catalog"]):
            raise ValueError("No catalog uri set")

    with ctx.obj["xnat_conn"] as session:
        xnat_to_FDP(session, config, catalog, fdpclient, sparqlclient=sparqlclient)


@cli_click.command(name="project")
@click.argument("project_id", default=None, type=str)
@click.option(
    "-o",
    "--output",
    "output",
    default=None,
    type=click.Path(writable=True, dir_okay=False),
    help="Destination file to write output to. If not set, the script will print serialized output to stdout.",
)
@click.option(
    "-f",
    "--format",
    default="turtle",
    type=click.Choice(
        ["xml", "n3", "turtle", "nt", "pretty-xml", "trix", "trig", "nquads", "json-ld", "hext"], case_sensitive=False
    ),
    help=(
        "The format that the output should be written in. This value references a"
        " Serializer plugin in RDFlib. Supportd values are: "
        ' "xml", "n3", "turtle", "nt", "pretty-xml", "trix", "trig", "nquads",'
        ' "json-ld" and "hext". Defaults to "turtle".'
    ),
)
@click.pass_context
def output_project(ctx: click.Context, project_id: str, output: click.Path, format: str):
    """Specify one project for DCAT extraction.

    The project is referred to by XNAT ID"""
    config = ctx.obj["config"]

    with ctx.obj["xnat_conn"] as session:
        dataset, uri = xnat_to_DCATDataset(session.projects[project_id], config)
        dataset_graph = dataset.to_graph(uri)

    if output:
        logger.debug("Output option set, serializing output to file %s in %s format", output, format)
        dataset_graph.serialize(destination=output, format=format)

    else:
        logger.debug("Sending output to stdout")
        print(dataset_graph.serialize(format=format))


if __name__ == "__main__":
    cli_click()
